# Generated by Django 2.0.9 on 2018-10-05 13:02

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0014_auto_20180918_1210'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='layer',
            options={'permissions': (('can_update_features_properties', 'Is able update geometries properties'),)},
        ),
    ]



from copy import deepcopy
from functools import reduce

from deepmerge import always_merger
try:
    from django.db.models import JSONField
except ImportError:  # TODO Remove when dropping Django releases < 3.1
    from django.contrib.postgres.fields import JSONField
from django.db import models
from django.utils.functional import cached_property

from geostore import GeometryTypes


class BaseUpdatableModel(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class LayerBasedModelMixin(BaseUpdatableModel):
    # Settings schema
    SETTINGS_DEFAULT = {
        'metadata': {
            'attribution': None,  # Json, eg. {'name': 'OSM contributors', href='http://openstreetmap.org'}
            'licence': None,  # String, eg. 'ODbL'
            'description': None,  # String
        },
        # Tilesets attributes
        'tiles': {
            'minzoom': 0,
            'maxzoom': 22,
            'pixel_buffer': 4,
            'features_filter': None,  # Json
            'properties_filter': None,  # Array of string
            'features_limit': 10000,
        }
    }
    settings = JSONField(default=dict, blank=True)
    geom_type = models.IntegerField(choices=GeometryTypes.choices(), null=True)

    @property
    def is_point(self):
        return self.layer_geometry in (GeometryTypes.Point,
                                       GeometryTypes.MultiPoint)

    @property
    def is_linestring(self):
        return self.layer_geometry in (GeometryTypes.LineString,
                                       GeometryTypes.MultiLineString)

    @property
    def is_polygon(self):
        return self.layer_geometry in (GeometryTypes.Polygon,
                                       GeometryTypes.MultiPolygon)

    @property
    def is_multi(self):
        return self.layer_geometry in (GeometryTypes.MultiPoint,
                                       GeometryTypes.MultiLineString,
                                       GeometryTypes.MultiPolygon)

    @cached_property
    def layer_geometry(self):
        """
        Return the geometry type of the layer using the first feature in
        the layer if the layer have no geom_type or the geom_type of the layer
        """
        if self.geom_type is None:
            feature = self.features.first()
            if feature:
                return feature.geom.geom_typeid
        return self.geom_type

    @cached_property
    def settings_with_default(self):
        return always_merger.merge(deepcopy(self.SETTINGS_DEFAULT), self.settings)

    def layer_settings(self, *json_path):
        """ Return the nested value of settings at path json_path.
            Raise an KeyError if not defined.
        """
        # Dives into settings using args
        return reduce(
            lambda a, v: a[v],  # Let raise KeyError on missing key
            json_path,
            self.settings) if self.settings is not None else None

    def layer_settings_with_default(self, *json_path):
        """ Return the nested value of settings with SETTINGS_DEFAULT as
            fallback at path json_path.
            Raise an KeyError if not defined.
        """
        # Dives into settings using args
        return reduce(
            lambda a, v: a[v],  # Let raise KeyError on missing key
            json_path,
            self.settings_with_default)

    def set_layer_settings(self, *json_path_value):
        """ Set last parameter as value at the path place into settings """

        json_path, value = json_path_value[:-1], json_path_value[-1]
        # Dive into settings until the last key of path,
        # and set the corresponding value
        settings = self.settings
        for key in json_path[:-1]:
            s = settings.get(key, {})
            settings[key] = s
            settings = s
        settings[json_path[-1]] = value

        try:
            # Delete the cached property
            del self.settings_with_default
        except AttributeError:
            pass  # Let's continue, cache was not set

    class Meta:
        abstract = True



# Generated by Django 3.0.2 on 2020-01-27 13:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0036_auto_20200116_0926'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='feature',
            index=models.Index(fields=['id', 'layer'], name='geostore_fe_id_bb7476_idx'),
        ),
        migrations.AddIndex(
            model_name='feature',
            index=models.Index(fields=['source', 'layer'], name='geostore_fe_source_cdd710_idx'),
        ),
        migrations.AddIndex(
            model_name='feature',
            index=models.Index(fields=['target', 'layer'], name='geostore_fe_target_b00bb7_idx'),
        ),
    ]



from django.core.management.base import CommandError

from geostore.models import Layer


class LayerCommandMixin(object):
    def _get_layer_by_pk(self, layer_pk):
        try:
            layer = Layer.objects.get(pk=layer_pk)
        except Layer.DoesNotExist:
            raise CommandError(f"Layer with pk {layer_pk} doesn't exist")
        return layer

    def _get_layer_by_name(self, name):
        try:
            return Layer.objects.get(name=name)
        except Layer.DoesNotExist:
            raise CommandError(f"Layer with name {name} doesn't exist")



import argparse
import json
from json.decoder import JSONDecodeError

from django.core.management.base import BaseCommand, CommandError
from django.utils.translation import gettext as _

from geostore.management.commands.mixins import LayerCommandMixin
from geostore.models import LayerGroup
from geostore.tiles.helpers import guess_maxzoom, guess_minzoom


class Command(LayerCommandMixin, BaseCommand):

    def add_arguments(self, parser):
        parser.add_argument('-pk', '--layer-pk',
                            type=int,
                            action="store",
                            required=True,
                            help=_("PK of the layer to update"))
        parser.add_argument('-l', '--layer',
                            action="store",
                            help=("Change the name of the layer"))
        parser.add_argument('-ls', '--layer_settings', nargs='?',
                            type=argparse.FileType('r'),
                            action="store",
                            help=("Replace JSON settings file to override default"))
        parser.add_argument('-gr', '--groups',
                            nargs='+',
                            help=_("Group names of the created layer"))
        parser.add_argument('-gz', '--guess_zoom',
                            action='store_true',
                            help=_("Guess min and max zoom from data"))

    def handle(self, *args, **options):
        layer_pk = options.get('layer_pk')
        layer = self._get_layer_by_pk(layer_pk)

        self._settings(layer, options)
        self._actions(layer, options)
        layer.save()

    def _settings(self, layer, options):
        layer_name = options.get('layer')
        group_names = options.get('groups')

        if group_names:
            layer.layer_groups.clear()
            for group_name in group_names:
                group, created = LayerGroup.objects.get_or_create(name=group_name)
                group.layers.add(layer)

        if layer_name:
            layer.name = layer_name

        layer_settings = options.get('layer_settings')
        if layer_settings:
            self._settings_settings(layer, layer_settings)

    def _settings_settings(self, layer, layer_settings):
        try:
            layer.settings = json.loads(layer_settings.read())
        except (JSONDecodeError, UnicodeDecodeError):
            raise CommandError("Please provide a valid layer settings file")

    def _actions(self, layer, options):
        guess_zoom = options.get('guess_zoom')

        if guess_zoom:
            layer.set_layer_settings('tiles', 'minzoom', guess_minzoom(layer))
            layer.set_layer_settings('tiles', 'maxzoom', guess_maxzoom(layer))



from django.db.models.signals import post_save
from django.dispatch import receiver

from geostore import settings as app_settings
from geostore.helpers import execute_async_func
from geostore.models import Feature, LayerRelation
from geostore.tasks import feature_update_relations_destinations, layer_relations_set_destinations


@receiver(post_save, sender=Feature)
def save_feature(sender, instance, **kwargs):
    if app_settings.GEOSTORE_RELATION_CELERY_ASYNC:
        execute_async_func(feature_update_relations_destinations, (instance.pk,))


@receiver(post_save, sender=LayerRelation)
def save_layer_relation(sender, instance, **kwargs):
    if app_settings.GEOSTORE_RELATION_CELERY_ASYNC:
        execute_async_func(layer_relations_set_destinations, (instance.pk,))



from enum import IntEnum

default_app_config = 'geostore.apps.GeostoreConfig'


class GeometryTypes(IntEnum):
    Point = 0
    LineString = 1
    # LinearRing 2
    Polygon = 3
    MultiPoint = 4
    MultiLineString = 5
    MultiPolygon = 6
    GeometryCollection = 7

    @classmethod
    def choices(cls):
        return [(geom_type.value, str(geom_type).split('.')[-1]) for geom_type in cls]

    @classmethod
    def shape_allowed_types(cls):
        """
        Types allowed in shapefile export
        """
        excluded = [GeometryTypes.GeometryCollection]
        return [geom_type for geom_type in cls if geom_type not in excluded]

    @classmethod
    def shape_allowed_type_names(cls):
        """
        Name types allowed in shapefile export
        """
        return [str(geom_type).split('.')[-1] for geom_type in cls.shape_allowed_types()]



# Generated by Django 2.0.13 on 2019-06-13 13:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0025_auto_20190613_1341'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='feature',
            index=models.Index(fields=['updated_at', 'layer'], name='geo_featu_updated_cfeac3_idx'),
        ),
    ]



import operator
try:
    from django.db.models import JSONField
except ImportError:  # TODO Remove when dropping Django releases < 3.1
    from django.contrib.postgres.fields import JSONField
from django.core.exceptions import FieldDoesNotExist
from django.db.models import Q
from rest_framework.filters import BaseFilterBackend, OrderingFilter, SearchFilter
from functools import reduce


class JSONFieldFilterBackend(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        query = Q()
        for param_name, param_value in request.query_params.items():
            try:
                field = (queryset.model
                                 ._meta
                                 .get_field(param_name.split('__')[0]))
            except FieldDoesNotExist:
                pass
            else:
                if isinstance(field, JSONField):
                    sub_query = Q(**{param_name: param_value})
                    try:
                        sub_query |= Q(**{param_name: int(param_value)})
                    except ValueError:
                        pass
                    query &= sub_query
        return queryset.filter(query)


class JSONFieldOrderingFilter(OrderingFilter):
    def get_valid_fields(self, queryset, view, context={}):
        fields = super().get_valid_fields(queryset, view, context=context)
        layer = view.get_layer()
        if layer and layer.schema:
            # allow ordering by property name ONLY if schema specified.
            # This prevents big queries on all layer's feature to find them
            for prop in layer.layer_properties:
                fields.append((f'properties__{prop}', layer.get_property_title(prop)))
        return fields


class JSONSearchField(SearchFilter):
    def get_search_fields(self, view, request):
        fields = []
        layer = view.get_layer()
        if layer and layer.schema:
            # allow search by property name ONLY if schema specified.
            # This prevents big queries on all layer's feature to find them
            for prop in layer.layer_properties:
                fields.append(f'properties__{prop}')
        return fields

    def filter_queryset(self, request, queryset, view):
        search_fields = self.get_search_fields(view, request)
        search_terms = self.get_search_terms(request)
        if search_terms:
            filters = list()
            for field in search_fields:
                filters.append(Q(**{f'{field}__icontains': search_terms[0]}))

            return queryset.filter(reduce(operator.or_, filters))
        else:
            return queryset



import json
import logging
import uuid

import fiona
from django.contrib.gis.geos import GEOSGeometry, GEOSException
from django.db import transaction
from fiona.transform import transform_geom

from geostore.import_export.helpers import ChunkIterator

from geostore import settings as app_settings
from geostore.tiles.decorators import zoom_update

ACCEPTED_PROJECTIONS = [
    'urn:ogc:def:crs:OGC:1.3:CRS84',
    'EPSG:4326',
]
logger = logging.getLogger(__name__)


class LayerImportMixin:
    def _fiona_shape_projection(self, shapefile):
        """ Return projection in EPSG format or raw Proj format extracted from
            shape
        """
        proj = shapefile.crs

        if not proj:
            raise GEOSException('Your shapefile does not have projection')

        if len(proj) == 1 or (len(proj) == 2 and proj.get('no_defs') is True):
            return proj.get('init')
        else:
            return fiona.crs.to_string(proj)

    def is_projection_allowed(self, projection):
        return projection in ACCEPTED_PROJECTIONS

    @zoom_update
    def from_shapefile(self, zipped_shapefile_file, id_field=None):
        """ Load ShapeFile content provided into a zipped archive.

        zipped_shapefile_file -- a file-like object on the zipped content
        id_field -- the field name used a identifier
        """
        with fiona.BytesCollection(zipped_shapefile_file.read()) as shape:
            # Extract source projection and compute if reprojection is required
            projection = self._fiona_shape_projection(shape)
            reproject = projection and not self.is_projection_allowed(projection.upper())

            for feature in shape:
                properties = {}
                for prop, value in feature.get('properties', {}).items():
                    try:
                        properties[prop] = json.loads(value)
                    except (json.JSONDecodeError, TypeError):
                        properties[prop] = value

                geometry = feature.get('geometry')

                if reproject:
                    geometry = transform_geom(
                        shape.crs,
                        f'EPSG:{app_settings.INTERNAL_GEOMETRY_SRID}',
                        geometry)
                identifier = properties.get(id_field, uuid.uuid4())

                self.features.create(
                    layer=self,
                    identifier=identifier,
                    properties=properties,
                    geom=GEOSGeometry(json.dumps(geometry)),
                )

    @zoom_update
    def from_geojson(self, geojson_data, id_field=None, update=False):
        """
        Import geojson raw data in a layer
        Args:
            geojson_data(str): must be raw text json data
        """

        geojson = json.loads(geojson_data)
        projection = geojson.get('crs', {}).get(
            'properties', {}).get('name', None)
        if projection and not self.is_projection_allowed(projection):
            raise GEOSException(
                f'GeoJSON projection {projection} must be in '
                f'{ACCEPTED_PROJECTIONS}')

        if update:
            self.features.all().delete()
        for feature in geojson.get('features', []):
            properties = feature.get('properties', {})
            identifier = properties.get(id_field, uuid.uuid4())
            self.features.update_or_create(
                layer=self,
                identifier=identifier,
                defaults={
                    'properties': properties,
                    'geom': GEOSGeometry(json.dumps(feature.get('geometry'))),
                }
            )

    def _initial_import_from_csv(self, chunks, options, operations):
        for chunk in chunks:
            entries = []
            for row in chunk:
                feature_args = {
                    "geom": None,
                    "properties": row,
                    "layer": self,
                }

                for operation in operations:
                    operation(feature_args, options)

                if not feature_args.get("geom"):
                    logger.warning('empty geometry,'
                                   f' row skipped : {row}')
                    continue
                entries.append(
                    self.features.model(**feature_args)
                )
            self.features.bulk_create(entries)

    def _complementary_import_from_csv(self, chunks, options, operations,
                                       pk_properties, fast=False):
        for chunk in chunks:
            sp = None
            if fast:
                sp = transaction.savepoint()
            for row in chunk:
                self._import_row_from_csv(row, pk_properties, operations,
                                          options)
            if sp:
                transaction.savepoint_commit(sp)

    def _import_row_from_csv(self, row, pk_properties, operations, options):
        feature_args = {
            "geom": None,
            "properties": row,
            "layer": self
        }
        for operation in operations:
            operation(feature_args, options)
        filter_kwargs = {
            f'properties__{p}': feature_args["properties"].get(p, '')
            for p in pk_properties}
        filter_kwargs['layer'] = feature_args.get("layer", self)
        if feature_args.get("geom"):
            self.features.update_or_create(
                defaults=feature_args,
                **filter_kwargs
            )
        else:
            self.features.filter(**filter_kwargs) \
                .update(properties=feature_args["properties"])

    @zoom_update
    def from_csv_dictreader(self, reader, pk_properties, options, operations,
                            init=False, chunk_size=1000, fast=False):
        """Import (create or update) features from csv.DictReader object
        :param reader: csv.DictReader object
        :param pk_properties: keys of row that is used to identify unicity
        :param init: allow to speed up import if there is only new Feature's
                    (no updates)
        :param chunk_size: only used if init=True, control the size of
                           bulk_create
        """
        chunks = ChunkIterator(reader, chunk_size)
        if init:
            self._initial_import_from_csv(
                chunks=chunks,
                options=options,
                operations=operations
            )
        else:
            self._complementary_import_from_csv(
                chunks=chunks,
                options=options,
                operations=operations,
                pk_properties=pk_properties,
                fast=fast
            )



# Generated by Django 3.0.6 on 2020-06-25 12:44

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0037_auto_20200624_1440'),
    ]

    operations = [
        migrations.AlterField(
            model_name='feature',
            name='layer',
            field=models.ForeignKey(db_index=False, on_delete=django.db.models.deletion.PROTECT, related_name='features', to='geostore.Layer'),
        ),
    ]



import jsonschema
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from jsonschema.validators import validator_for


def validate_json_schema(value):
    """
    Validate json schema
    """
    try:
        if value:
            # check only if schema defined
            cls = validator_for(value)
            cls.check_schema(value)
    except Exception as e:
        raise ValidationError(message=e.message)

    return value


def validate_json_schema_data(value, schema):
    """
    Validate data according json schema
    """
    try:
        # check result schema
        if value and schema:
            properties = schema.get('properties').keys()
            unexpected_properties = value.keys() - properties
            if unexpected_properties:
                # value key(s) not in expected properties
                raise ValidationError(message=_(f"{unexpected_properties} not in schema properties"))
            jsonschema.validate(value, schema)
    except jsonschema.exceptions.ValidationError as e:
        raise ValidationError(message=e.message)

    return value


def validate_geom_type(layer_geom_type_id, feature_geom_type_id):
    if isinstance(layer_geom_type_id, int):
        if layer_geom_type_id != feature_geom_type_id:
            raise ValidationError(message=_('Geometry type is not the same on the layer'))
    return feature_geom_type_id


def validate_geom(feature_geom):
    if feature_geom.empty:
        raise ValidationError(_('Geometry is empty'))
    if not feature_geom.valid:
        raise ValidationError(_('Geometry is not valid'))
    return feature_geom



import glob
import json
import os
from tempfile import TemporaryDirectory

import fiona
from django.core.serializers import serialize
from fiona.crs import from_epsg

from geostore import GeometryTypes
from geostore import settings as app_settings
from geostore.import_export.helpers import make_zipfile_bytesio, get_serialized_properties
from geostore.renderers import KMLRenderer


class LayerExportMixin:
    def to_geojson(self):
        if not self.features.count():
            return
        return serialize('geojson',
                         self.features.all(),
                         fields=('properties',),
                         geometry_field='geom',
                         properties_field='properties')

    def to_shapefile(self):
        if not self.features.count():
            return
        with TemporaryDirectory() as shape_folder:
            shapes = {}
            # get all accepted types if geom_type not defined, else keep selected
            type_to_check = GeometryTypes.shape_allowed_type_names() \
                if not self.geom_type else \
                [GeometryTypes(self.geom_type).name]

            # Create one shapefile by kind of geometry
            for geom_type in type_to_check:
                schema = {
                    'geometry': geom_type,
                    'properties': self.layer_properties,
                }

                shapes[geom_type] = fiona.open(
                    shape_folder,
                    layer=geom_type,
                    mode='w',
                    driver='ESRI Shapefile',
                    schema=schema,
                    encoding='UTF-8',
                    crs=from_epsg(app_settings.INTERNAL_GEOMETRY_SRID)
                )

            # Export features to each kind of geometry
            for feature in self.features.all():
                shapes[feature.geom.geom_type].write({
                    'geometry': json.loads(feature.geom.json),
                    'properties': get_serialized_properties(self, feature.properties)
                })

            # Close fiona files
            for geom_type, shape in shapes.items():
                shape_size = len(shape)
                shape.close()

                # Delete empty shapes
                if not shape_size:
                    for filename in glob.iglob(os.path.join(shape_folder, f'{geom_type}.*')):
                        os.remove(filename)
            # Zip to BytesIO and return shape files
            return make_zipfile_bytesio(shape_folder)

    def to_kml(self):
        from geostore.serializers import FeatureSerializer
        if not self.features.count():
            return
        return KMLRenderer().render(FeatureSerializer(self.features.all(), many=True).data)



import argparse
import csv
import sys

from django.core.management import BaseCommand
from django.utils.module_loading import import_string
from django.utils.translation import gettext as _

from geostore.models import Layer


class Command(BaseCommand):
    help = _('Import INSEE data from csv to db.')

    def add_arguments(self, parser):
        parser.add_argument('-b', '--bulk',
                            action='store_true',
                            dest='bulk',
                            default=False,
                            help=_("Delete all stored INSEE data,"
                                   " DO NOT USE IN PROD.")
                            )
        parser.add_argument('-c', '--creations-per-transaction',
                            dest='chunk_size',
                            type=int,
                            default=1000,
                            help=_('Number of rows per transaction')
                            )
        parser.add_argument('-cd', '--delimiter',
                            action='store',
                            dest='delimiter',
                            default=';',
                            required=False,
                            help=_('Specify CSV delimiter')
                            )
        parser.add_argument('-cq', '--quotechar',
                            action='store',
                            dest='quotechar',
                            default='"',
                            required=False,
                            help=_('Specify CSV quotechar')
                            )
        parser.add_argument('-cs', '--source',
                            dest='source',
                            type=argparse.FileType('r',
                                                   encoding='iso-8859-15'),
                            default=sys.stdin,
                            required=True,
                            help=_('Specify CSV path'),
                            )
        parser.add_argument('-f', '--fast',
                            action='store_true',
                            default=False,
                            help=_("If present and it's not an initial import"
                                   " will speed up features creation. But no"
                                   " rollback is possible. If something broke"
                                   " up during import, the import will stop "
                                   " with half data in database.")
                            )
        parser.add_argument('-i', '--init',
                            action='store_true',
                            dest='init',
                            default=False,
                            help=_('Improve performance of initial import')
                            )
        parser.add_argument('-k', '--key',
                            required=True,
                            action='append',
                            dest='pk_properties',
                            help=_("Define primary keys of this data."
                                   " Example with companies:"
                                   " --key=SIREN --key=NIC"
                                   " Example with INSEE:"
                                   " --key=code_insee")
                            )
        parser.add_argument('-l', '--layer',
                            required=True,
                            action='store',
                            dest='layer',
                            help=_('Specify layer name')
                            )
        parser.add_argument('-o', '--operation',
                            required=True,
                            action='append',
                            dest='operations',
                            help=_('Specify transform functions')
                            )
        parser.add_argument('--longitude',
                            required=False,
                            action='store',
                            dest='longitude',
                            help=_('Name of longitude column')
                            )
        parser.add_argument('--latitude',
                            required=False,
                            action='store',
                            dest='latitude',
                            help=_('Name of latitude column')
                            )

    def handle(self, *args, **options):
        layer_name = options.get('layer')
        layer = Layer.objects.get_or_create(name=layer_name)[0]

        if options['bulk']:
            layer.features.all().delete()

        operations = None
        if options['operations']:
            operations = [import_string(path) for path in
                          options.get('operations')]
        reader = csv.DictReader(options.get('source'),
                                delimiter=options.get('delimiter'),
                                quotechar=options.get('quotechar'))

        layer.from_csv_dictreader(
            reader=reader,
            options=options,
            operations=operations,
            pk_properties=options.get('pk_properties'),
            init=options.get('init'),
            chunk_size=options.get('chunk_size'),
            fast=options.get('fast')
        )



import io
import json
import logging
import os
import zipfile

from django.contrib.gis.geos import Point
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django.core.mail import send_mail
from django.template.loader import get_template
from django.utils.timezone import now
from django.utils.translation import gettext as _

logger = logging.getLogger(__name__)


def make_zipfile_bytesio(base_dir):
    zip_file = io.BytesIO()

    with zipfile.ZipFile(zip_file, "w",
                         compression=zipfile.ZIP_DEFLATED) as zf:
        path = os.path.normpath(base_dir)
        if path != os.curdir and path != base_dir:
            zf.write(path, os.path.relpath(path, base_dir))
            logger.info("adding '%s'", path)
        for dirpath, dirnames, filenames in os.walk(base_dir):
            for name in sorted(dirnames):
                path = os.path.normpath(os.path.join(dirpath, name))
                zf.write(path, os.path.relpath(path, base_dir))
                logger.info("adding '%s'", path)
            for name in filenames:
                path = os.path.normpath(os.path.join(dirpath, name))
                if os.path.isfile(path):
                    zf.write(path, os.path.relpath(path, base_dir))
                    logger.info("adding '%s'", path)

    return zip_file


def send_mail_export(user, path=None):
    url = default_storage.url(path)
    context = {"username": user.get_username(), "url": url}
    if not path:
        template_email = 'exports_no_data'
    else:
        template_email = 'exports'
    html = get_template('geostore/emails/{}.html'.format(template_email))
    html_content = html.render(context)
    txt = get_template('geostore/emails/{}.txt'.format(template_email))
    txt_content = txt.render(context)
    send_mail(_('Your data export is ready'), txt_content, None, [getattr(user, user.get_email_field_name())],
              html_message=html_content, fail_silently=True)


def save_generated_file(user_id, layer_name, format_file, string_file):
    string_file = string_file.encode('utf-8') if not isinstance(string_file, bytes) else string_file
    path = default_storage.save('exports/users/{}/{}_{}.{}'.format(user_id,
                                                                   layer_name,
                                                                   int(now().timestamp()),
                                                                   format_file),
                                ContentFile(string_file))
    return path


class ChunkIterator:
    def __init__(self, iterator, chunksize):
        self.iterator = iterator
        self.chunksize = chunksize

    def __iter__(self):
        return self

    def __next__(self):
        chunk = []
        try:
            for i in range(self.chunksize):
                chunk.append(next(self.iterator))
        finally:
            if chunk:
                return chunk
            else:
                raise StopIteration

    def next(self):
        return self.__next__()


def get_serialized_properties(layer, feature_properties):
    properties = {k: None for k in layer.layer_properties}
    for prop, value in feature_properties.items():
        if isinstance(value, str):
            properties[prop] = value
        else:
            properties[prop] = json.dumps(value)
    return properties


class GeometryDefiner:
    LONGITUDE = 'longitude'
    LATITUDE = 'latitude'

    @staticmethod
    def get_geometry(column_names, row):
        if not isinstance(column_names, dict):
            return None
        if sorted(column_names.keys()) == [GeometryDefiner.LATITUDE,
                                           GeometryDefiner.LONGITUDE]:
            lat_column = column_names.get(GeometryDefiner.LATITUDE)
            long_column = column_names.get(GeometryDefiner.LONGITUDE)
            if all(row.get(column) for column in [long_column, lat_column]):
                x = float(row.get(long_column))
                y = float(row.get(lat_column))
                return Point(x, y)
        return None



import os
import subprocess
import tempfile
import uuid
from xml.etree import ElementTree as ET

import requests
from django.core.management.base import BaseCommand, CommandError
from django.utils.translation import gettext as _

from geostore.management.commands.mixins import LayerCommandMixin
from geostore.models import Layer


class Command(LayerCommandMixin, BaseCommand):
    overpass_url = "http://overpass-api.de/api/interpreter"

    def add_arguments(self, parser):
        parser.add_argument('query', action="store",
                            help=_("Overpass query without type (out:json)"
                                   ))
        parser.add_argument('-t', '--type',
                            required=True,
                            action='store',
                            dest='type',
                            help=_('Specify type of layer'
                                   ))
        parser.add_argument('-l', '--layer',
                            action="store",
                            help=_("Name of created layer "
                                   "containing GeoJSON data. "
                                   "If not provided an uuid4 is set."
                                   ))
        parser.add_argument('-pk', '--layer-pk',
                            type=int,
                            action="store",
                            help=_("PK of the layer where to insert "
                                   "the features.\n"
                                   "A new layer is created if not "
                                   "present."
                                   ))
        parser.add_argument('-gr', '--group',
                            action="store",
                            default="__nogroup__",
                            help=_("Group name of the created layer"
                                   ))
        parser.add_argument('-i', '--identifier',
                            action="store",
                            help=_("Field in properties that will be used as "
                                   "identifier of the features, so features"
                                   " can be grouped on layer's operations"
                                   ))

    def handle(self, *args, **options):
        query = options.get('query')
        layer_name = options.get('layer') or uuid.uuid4()
        type_features = options.get('type')
        layer_pk = options.get('layer_pk')
        identifier = options.get('identifier')
        verbosity = options.get('verbosity')
        response = requests.get(self.overpass_url,
                                params={'data': query})
        try:
            ET.fromstring(response.content)
        except ET.ParseError:
            if response.status_code != 400:
                raise CommandError("Overpass didn't give any information")
            raise CommandError('The query is not valid')

        value, log_error = self.launch_cmd_ogr2ogr(response.content, type_features)

        if verbosity >= 1:
            self.stderr.write(log_error)
        if not value:
            raise CommandError('Ogr2ogr failed to create the geojson')
        if layer_pk:
            layer = self._get_layer_by_pk(layer_pk)
        else:
            settings = {
                'metadata': {
                    'attribution': '<a href=\'http://openstreetmap.org\'>OSM contributors</a>',
                    'licence': 'ODbL',
                }
            }
            layer = Layer.objects.create(name=layer_name, settings=settings)
        layer.from_geojson(value, identifier)

    def launch_cmd_ogr2ogr(self, content, type_features):
        tmp_osm = tempfile.NamedTemporaryFile(mode='w+b', delete=False)
        tmp_osm.write(content)
        tmp_osm.close()
        try:
            proc = subprocess.run(
                args=[
                    'ogr2ogr',
                    '-f', 'GeoJSON', '/vsistdout/',
                    tmp_osm.name,
                    type_features,
                    '--config', 'OSM_USE_CUSTOM_INDEXING', 'NO',
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                encoding='utf8',)
            value = proc.stdout
            log_error = proc.stderr
        except subprocess.CalledProcessError:
            raise CommandError("Command ogr2ogr failed")
        finally:
            os.unlink(tmp_osm.name)
        return value, log_error



# Generated by Django 2.0.13 on 2019-05-23 13:28

from django.db import migrations, models
import geostore.models


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0023_layer_type_geom'),
    ]

    operations = [
        migrations.AlterField(
            model_name='layer',
            name='geom_type',
            field=models.IntegerField(choices=[(0, geostore.GeometryTypes(0)), (1, geostore.GeometryTypes(1)), (3, geostore.GeometryTypes(3)), (4, geostore.GeometryTypes(4)), (5, geostore.GeometryTypes(5)), (6, geostore.GeometryTypes(6)), (7, geostore.GeometryTypes(7))], null=True),
        ),
    ]



from celery import shared_task
from django.contrib.auth import get_user_model

from geostore.import_export.helpers import save_generated_file, send_mail_export
from geostore.models import Feature, LayerRelation, Layer


@shared_task
def feature_update_relations_destinations(feature_id, relation_id=None):
    """ Update all feature layer relations as origin """
    feature = Feature.objects.get(pk=feature_id)
    feature.sync_relations(relation_id)

    return True


@shared_task
def layer_relations_set_destinations(relation_id):
    """ Update all feature layer as origin for a relation """
    relation = LayerRelation.objects.get(pk=relation_id)

    for feature_id in relation.origin.features.values_list('pk', flat=True):
        feature_update_relations_destinations.delay(feature_id, relation_id)

    return True


@shared_task
def generate_shapefile_async(layer_id, user_id):
    layer = Layer.objects.get(pk=layer_id)
    user = get_user_model().objects.get(pk=user_id)
    file = layer.to_shapefile()

    path = save_generated_file(user_id, layer.name, 'zip', file.getvalue()) if file else None
    send_mail_export(user, path)


@shared_task
def generate_geojson_async(layer_id, user_id):
    layer = Layer.objects.get(pk=layer_id)
    user = get_user_model().objects.get(pk=user_id)

    file = layer.to_geojson()

    path = save_generated_file(user_id, layer.name, 'geojson', file) if file else None
    send_mail_export(user, path)


@shared_task
def generate_kml_async(layer_id, user_id):
    layer = Layer.objects.get(pk=layer_id)
    user = get_user_model().objects.get(pk=user_id)

    file = layer.to_kml()
    path = save_generated_file(user_id, layer.name, 'kml', file) if file else None
    send_mail_export(user, path)



# Generated by Django 3.2 on 2021-04-29 08:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0044_auto_20201106_1638'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='featureextrageom',
            constraint=models.CheckConstraint(check=models.Q(geom__isvalid=True), name='geom_extra_is_valid'),
        ),
        migrations.AddConstraint(
            model_name='featureextrageom',
            constraint=models.CheckConstraint(check=models.Q(geom__isempty=False), name='geom_extra_is_empty'),
        ),
    ]



# Generated by Django 2.2.5 on 2019-09-26 08:03

import django.contrib.postgres.indexes
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0028_remove_layer_group'),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='feature',
            name='geo_featu_identif_212b28_idx',
        ),
        migrations.RemoveIndex(
            model_name='feature',
            name='geo_featu_layer_i_3dcdde_gist',
        ),
        migrations.RemoveIndex(
            model_name='feature',
            name='geo_feat_layer_i_921f95_idx',
        ),
        migrations.RemoveIndex(
            model_name='feature',
            name='geo_featu_updated_62a19a_idx',
        ),
        migrations.RemoveIndex(
            model_name='feature',
            name='geo_featu_updated_cfeac3_idx',
        ),
        migrations.AddIndex(
            model_name='feature',
            index=models.Index(fields=['layer'], name='geostore_fe_layer_i_71cad6_idx'),
        ),
        migrations.AddIndex(
            model_name='feature',
            index=models.Index(fields=['updated_at'], name='geostore_fe_updated_fcdc6b_idx'),
        ),
        migrations.AddIndex(
            model_name='feature',
            index=models.Index(fields=['updated_at', 'layer'], name='geostore_fe_updated_bdb823_idx'),
        ),
        migrations.AddIndex(
            model_name='feature',
            index=models.Index(fields=['identifier'], name='geostore_fe_identif_f6c552_idx'),
        ),
        migrations.AddIndex(
            model_name='feature',
            index=django.contrib.postgres.indexes.GistIndex(fields=['layer', 'geom'], name='geostore_fe_layer_i_c3168f_gist'),
        ),
    ]



# Generated by Django 3.1.1 on 2020-10-08 12:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0041_auto_20200625_1515'),
    ]

    operations = [
        migrations.AddField(
            model_name='layer',
            name='routable',
            field=models.BooleanField(default=False, help_text='Used for make layer routable'),
        ),
    ]



# Generated by Django 3.2.7 on 2021-10-13 13:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0045_auto_20210429_0818'),
    ]

    operations = [
        migrations.AlterField(
            model_name='layer',
            name='geom_type',
            field=models.IntegerField(choices=[(0, 'Point'), (1, 'LineString'), (3, 'Polygon'), (4, 'MultiPoint'), (5, 'MultiLineString'), (6, 'MultiPolygon'), (7, 'GeometryCollection')], null=True),
        ),
        migrations.AlterField(
            model_name='layerextrageom',
            name='geom_type',
            field=models.IntegerField(choices=[(0, 'Point'), (1, 'LineString'), (3, 'Polygon'), (4, 'MultiPoint'), (5, 'MultiLineString'), (6, 'MultiPolygon'), (7, 'GeometryCollection')], null=True),
        ),
    ]
